@@ -1,7 +1,10 @@
-## ----setup, include=FALSE-----------------------------------------------------
-knitr::opts_chunk$set(echo = TRUE)
+## ----include = FALSE----------------------------------------------------------
+knitr::opts_chunk$set(
+  collapse = TRUE,
+  comment = "#>"
+)
 
-## -----------------------------------------------------------------------------
+## ----setup--------------------------------------------------------------------
 library(lrstat)
 
 ## -----------------------------------------------------------------------------
@@ -27,6 +30,6 @@ lrsim(kMax = 2, criticalValues = c(2.250, 2.025),
       gamma1 = -log(1-0.05)/12, 
       gamma2 = -log(1-0.05)/12,
       accrualDuration = 22,
-      plannedEvents = c(253, 316), 
+      plannedEvents = c(252, 315), 
       maxNumberOfIterations = 10000, seed = 314159)
 
